// MSP3520 - ILI9488/XPT2046 interface routines

#include <xc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "msp3520.h"
#include "font.h"

// microsecond wait
void
wait_us( int t )
{
  for( ; t; t -- ) NOP();
}

// milisecond wait
void
wait_ms( int t )
{
  for( ; t; t -- ) wait_us( 91 );
}

// software SPI
unsigned char
software_spi8( unsigned char d )
{
  int i;
  unsigned char r;

  r = 0;
  for( i = 0; i < 8; i ++ ) {
    if( d & 0x80 ) MOSI = 1;
    else           MOSI = 0;
    d <<= 1;
    SCK = 1;
    r <<= 1;
    if( MISO ) r |= 0x01;
    SCK = 0;
  }
  return r;
}

void
iniz_spi1( void )
{
  TRISBbits.TRISB5 = 1; // MISO1
  TRISBbits.TRISB4 = 0; // SCK1
  TRISCbits.TRISC7 = 0; // MOSI1
  TRISCbits.TRISC6 = 0; // LCDCS
  TRISCbits.TRISC5 = 0; // LCDRS

  SSP1CON1 = 0b00000000;  // status/control register reset
  SSP1STAT = 0b00000000;
  SSP1CON1bits.SSPM = 0b0000; // SPI master mode fosc/4 clock
  SSP1STATbits.CKE = 1; // SCK phase
  SSP1CON1bits.SSPEN = 1; // SPI enable
}

unsigned char
hardware_spi8( unsigned char d )
{
  unsigned char r;

  while( 0 != SSP1STATbits.BF ) ; // wait for buffer empty
  SSP1BUF = d;
  while( 0 == SSP1STATbits.BF ) ; //wait for transmit start / receive buffer full
  r = SSP1BUF;
//  wait_us( 1 );
  return r;
}

// MSP3520(ILI9488) 8 color mode
unsigned int lcdx, lcdy;    // XY position, graphic/character current point
unsigned short int fcolor = 0x0007, bcolor = 0x0000;  // character color
unsigned int fcache, cachex, caxhey; // flame memory cache
// color is xxRGBRGB 3bit color, 1-1-1 bit RGB
unsigned short int color_code[10] = {
  0b1100000100000000, // brown
  0b1111100000000000, // red
  0b1111101000000000, // orange
  0b1111111111100000, // yellow
  0b0000011111100000, // green
  0b0000000000011111, // blue
  0b1111100000011111, // purple
  0b0100001000001000, // gray
  0b1111111111111111, // white
  0b0000000000000000, // black
};

// send commnand to ILI9488
void
command_ili9488( unsigned char c )
{
  LCDRS = 0; // select command register
  SPI8( c );
}

// send data to ILI9488
unsigned char
data_ili9488( unsigned char c )
{
  LCDRS = 1; // select data register
  return SPI8( c );
}

// set draw resion to ILI9488
void
address_set_ili9488( unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2 )
{
  LCDCS = 0;
  command_ili9488( 0x2A );  // column address set
  LCDRS = 1;
  SPI8( x1 >> 8 );
  SPI8( x1 & 0xFF );
  SPI8( x2 >> 8 );
  SPI8( x2 & 0xFF );
  command_ili9488( 0x2B ); // page address set
  LCDRS = 1;
  SPI8( y1 >> 8 );
  SPI8( y1 & 0xFF );
  SPI8( y2 >> 8 );
  SPI8( y2 & 0xFF );
  LCDCS = 1;
}

// initialize ILI9488 LCD controller and display
void
iniz_ili9488( void )
{
//  RESET = 0;
//  wait_ms( 15 );
//  RESET = 1;
  wait_ms( 100 );

  LCDCS = 0;
  command_ili9488( 0xF7 );  // Adjust control 3
  data_ili9488( 0xA9 );
  data_ili9488( 0x51 );
  data_ili9488( 0x2C );
  data_ili9488( 0x82 );     // DSI write DCS command use loose packet RGB666
  //                        // 3 bytes/pixel rrrrrrxx ggggggxx bbbbbbxx

  command_ili9488( 0xC0 );  // Power control 1
  data_ili9488( 0x11 );
  data_ili9488( 0x09 );

  command_ili9488( 0xC1 );  // Power control 2
  data_ili9488( 0x41 );

  command_ili9488( 0xC5 );  // VCOM control 1
  data_ili9488( 0x00 );
  data_ili9488( 0x0A );
  data_ili9488( 0x80 );

  command_ili9488( 0xB1 );  // Flame rate control
  data_ili9488( 0xB0 );
  data_ili9488( 0x11 );

  command_ili9488( 0xB4 );  // Display inversion control
  data_ili9488( 0x02 );

  command_ili9488( 0xB6 );  // Display function control
  data_ili9488( 0x02 );
  data_ili9488( 0x22 );

  command_ili9488( 0xB7 );  // Entry mode set
  data_ili9488( 0xC6 );

  command_ili9488( 0xBE );  // HS lanes control
  data_ili9488( 0x00 );
  data_ili9488( 0x04 );

  command_ili9488( 0xE9 );  // Set Image function
  data_ili9488( 0x00 );
  
  command_ili9488( 0x36 );  // Memory access control 
  data_ili9488( 0x08 );

  command_ili9488( 0x3A );  // Interface Pixel format
  data_ili9488( 0x61 );     // MPU interface format 3bits/Pixel(8color)

  command_ili9488( 0xE0 );  // positive gamma control
  data_ili9488( 0x00 );
  data_ili9488( 0x07 );
  data_ili9488( 0x10 );
  data_ili9488( 0x09 );
  data_ili9488( 0x17 );
  data_ili9488( 0x0B );
  data_ili9488( 0x41 );
  data_ili9488( 0x89 );
  data_ili9488( 0x4B );
  data_ili9488( 0x0A );
  data_ili9488( 0x0C );
  data_ili9488( 0x0E );
  data_ili9488( 0x18 );
  data_ili9488( 0x1B );
  data_ili9488( 0x0F );

  command_ili9488( 0xE1 );  // Negative gamma control
  data_ili9488( 0x00 );
  data_ili9488( 0x17 );
  data_ili9488( 0x1A );
  data_ili9488( 0x04 );
  data_ili9488( 0x0E );
  data_ili9488( 0x06 );
  data_ili9488( 0x2F );
  data_ili9488( 0x45 );
  data_ili9488( 0x43 );
  data_ili9488( 0x02 );
  data_ili9488( 0x0A );
  data_ili9488( 0x09 );
  data_ili9488( 0x32 );
  data_ili9488( 0x36 );
  data_ili9488( 0x0F );

  command_ili9488( 0x11 );  // exit sleep mode
  wait_ms( 120 );
  command_ili9488( 0x29 );  // display on
  LCDCS = 1;
  lcdx = lcdy = 0;
  fcache = cachex = caxhey = 0xFFFF; // no data
  fcolor = 0xFFFF;
  bcolor = 0x0000;
  wait_us( 20 );
}

void
display_on_ili9488( void )
{
  LCDCS = 0;
  command_ili9488( 0x29 );  // display on
  LCDCS = 1;
}


// fill LCD screen with color - clear screen
int cls_flag = 0;
void
fill_ili9488( unsigned short int c )
{
  unsigned int x, y;
  
  if( 0 != cls_flag ) return;
  LCDCS = 0;
  command_ili9488( 0x3A );  // Interface Pixel format
  data_ili9488( 0x61 );     // MPU interface format 3bits/Pixel(8color)
  LCDCS = 1;
  wait_us( 1 );
  address_set_ili9488( 0, 0, YSIZE, XSIZE );
  LCDCS = 0;
  command_ili9488( 0x2C ); // memory write
  LCDRS = 1;

  for( x = 0; x < YSIZE; x ++ ) {
    for( y = 0; y < XSIZE / 2; y ++ ) {
      SPI8( c );
    }
  }
  LCDCS = 1;
  lcdx = lcdy = 0;
  fcache = cachex = caxhey = 0xFFFF; // no data
  bcolor = c;
  wait_us( 1 );
  LCDCS = 0;
  command_ili9488( 0x3A );  // Interface Pixel format
  data_ili9488( 0x66 );     // MPU interface format 3bits/Pixel(8color)
  LCDCS = 1;
  cls_flag = 1;
}

// ILI9488 / MSP5320 Coordinate system
// (0,319)                        (479,319)
//   +---------------------------------+
//   |                                 |
//   |                                 |
//   |                                 |
//   |                                 |
//   +---------------------------------+
// (0,0)                           (479,0)

// vertical line
void
vertical_line_ili9488( unsigned int x, unsigned int y, unsigned int l, unsigned short int c )
{
  unsigned int i;

  address_set_ili9488( y, x, y + l, x ); // set draw area
  LCDCS = 0;
  command_ili9488( 0x2C ); // memory write
  LCDRS = 1;
  for( i = 0; i < l; i ++ ) {
    SPI8( ( c >> 8 ) & 0xF8 );
    SPI8( ( c >> 3 ) & 0xFC );
    SPI8( ( c << 3 ) & 0xFF );
  }
  LCDCS = 1;
  lcdx = x;
  lcdy = y + l;
  cls_flag = 0;
}

// horizontal line
void
horizontal_line_ili9488( unsigned int x, unsigned int y, unsigned int l, unsigned short int c )
{
  unsigned int i;

  address_set_ili9488( y, x, y, x + l ); // set draw area
  LCDCS = 0;
  command_ili9488( 0x2C ); // memory write
  LCDRS = 1;
  for( i = 0; i < l; i ++ ) {
    SPI8( ( c >> 8 ) & 0xF8 );
    SPI8( ( c >> 3 ) & 0xFC );
    SPI8( ( c << 3 ) & 0xFF );
  }
  LCDCS = 1;
  lcdx = x + l;
  lcdy = y;
  cls_flag = 0;
}

// set pont
void
pset_ili9488( unsigned int x, unsigned int y, unsigned short int c )
{
  address_set_ili9488( y, x, y +1, x );  // set draw address
  LCDCS = 0;
  command_ili9488( 0x2C ); // memory write
  LCDRS = 1;
  SPI8( ( c >> 8 ) & 0xF8 );
  SPI8( ( c >> 3 ) & 0xFC );
  SPI8( ( c << 3 ) & 0xFF );
  LCDCS = 1;
  lcdx = x;
  lcdy = y;
  cls_flag = 0;
}

// draw start point set
void
locate_ili9488( unsigned int x, unsigned int y )
{
  lcdx = x; // draw posision set
  lcdy = y;
  cls_flag = 0;
}

// put a character ( 8x16 font )
void
putchar_ili9488( char c )
{
  unsigned short int i, j, f;

  for( j = 0; j < 8; j ++ ) {
    f = font8x16[ ( c - 0x20 ) * 8 + j ];
    f = ( f >> 8 ) | ( ( f & 0xff ) << 8 );
    address_set_ili9488( lcdy, lcdx + j, lcdy + 16, lcdx + j );
    LCDCS = 0;
    command_ili9488( 0x2C ); // memory write
    LCDRS = 1;
    for( i = 0; i < 16; i ++ ) {
      if( f & 0x8000 ) {
        SPI8( ( fcolor >> 8 ) & 0xF8 );
        SPI8( ( fcolor >> 3 ) & 0xFC );
        SPI8( ( fcolor << 3 ) & 0xFF );
      } else {
        SPI8( ( bcolor >> 8 ) & 0xF8 );
        SPI8( ( bcolor >> 3 ) & 0xFC );
        SPI8( ( bcolor << 3 ) & 0xFF );
      }
      f <<= 1;
    }
  wait_us( 1 );
  }
  lcdx += 8;
  if( 480 <= lcdx ) { 
    lcdx = 0;
    lcdy -= 20;
  }
  cls_flag = 0;
}

// put string
void
puts_ili9488( char *s )
{
  while( *s ) putchar_ili9488( *s++ );
  cls_flag = 0;
}

// draw line(DDA)
void
line_ili9488( unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2, unsigned short int c )
{
  int x, y;
  int dx, dy;
  int e;

  dx = x2 - x1;
  dy = y2 - y1;
  e = 0; x = y = 0; 
  if( dx >= dy ) {
    for( x = 0; x <= dx; x ++ ) {
      pset_ili9488( x1 + x, y1 + y, c );
      e += dy;
      if( 2 * e >= dx ) {
        y++;
        e -= dx;
      }
      else if( -( 2 * e ) >= dx ) {
        y--;
        e += dx;
      }
    }
  } else {
    for( y = 0; y <= dy; y ++ ) {
      pset_ili9488( x1 + x, y1 + y, c );
      e += dx;
      if( 2 * e >= dy ) {
        x++;
        e -= dy;
      }
      else if( -( 2 * e ) >= dy ) {
        x--;
        e += dy;
      }
    }
  }
  lcdx = x2;
  lcdy = y2;
  cls_flag = 0;
}

// draw rectangle(fill) - clear area
//  must X1 < X2 & Y1 < Y2
void
rectangle_ili9488( unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2, unsigned short int c )
{
  unsigned int l, y;

  if( x1 >= x2 || y1 >= y2 ) return;
  l = x2 - x1;
  for( y = y1; y <= y2; y++ ) {
    horizontal_line_ili9488( x1, y, l, c );
  }
  cls_flag = 0;
}

// put a symbol ( 8x16 font )
void
put_symbol_ili9488( char c, int mag )
{
  unsigned short int i, j, f;
  int sx, sy;

  if( 1 > mag || 10 < mag || 0x20 > c ) return;
  sx = lcdx; sy = lcdy;
  for( j = 0; j < 8; j ++ ) {
    f = font8x16[ ( c - 0x20 ) * 8 + j ];
    f = ( f >> 8 ) | ( ( f & 0xff ) << 8 );
    for( i = 0; i < 16; i ++ ) {
      if( f & 0x8000 ) {
        rectangle_ili9488( sx + j * mag, sy + i * mag, 
                           sx + ( j + 1 )* mag - 1, sy + ( i + 1 )* mag - 1, fcolor );
      } else {
        rectangle_ili9488( sx + j * mag, sy + i * mag, 
                           sx + ( j + 1 )* mag - 1, sy + ( i + 1 )* mag - 1, bcolor );
      }
      f <<= 1;
    }
  }
  display_on_ili9488();
  lcdx = sx + mag * 8;
  lcdy = sy;
  if( XSIZE <= lcdx ) lcdx = 0;
  cls_flag = 0;
}

// put symbol string
void
puts_symbol_ili9488( int mag, char *s )
{
  while( *s ) put_symbol_ili9488( *s++, mag );
  cls_flag = 0;
}

